from datetime import datetime

from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.conf import settings
from django.test import TestCase as DjangoTestCase

from django_comments_tree.models import (TreeComment, CommentAssociation,
                                         MaxThreadLevelExceededException)
from django_comments_tree.tests.models import Article, Diary


class ArticleBaseTestCase(DjangoTestCase):
    def setUp(self):
        self.article_1 = Article.objects.create(
            title="September", slug="september", body="During September...")
        self.article_2 = Article.objects.create(
            title="October", slug="october", body="What I did on October...")


class TreeCommentManagerTestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        self.article_ct = ContentType.objects.get(app_label="tests",
                                                  model="article")

        self.site1 = Site.objects.get(pk=1)
        self.site2 = Site.objects.create(domain='site2.com', name='site2.com')

        self.root_1 = TreeComment.objects.get_or_create_root(self.article_1)
        self.root_1_pk = self.root_1.pk
        self.root_2 = TreeComment.objects.get_or_create_root(self.article_1, site=self.site2)
        self.root_2_pk = self.root_2.pk

    def post_comment_1(self):
        r = TreeComment.objects.get(pk=self.root_1_pk)
        r.add_child(comment="just a testing comment")

    def post_comment_2(self):
        r = TreeComment.objects.get(pk=self.root_2_pk)
        r.add_child(comment="yet another comment")

    def post_comment_3(self):
        r = TreeComment.objects.get(pk=self.root_1_pk)
        r.add_child(comment="and another one")

    def post_comment_4(self):
        r = TreeComment.objects.get(pk=self.root_2_pk)
        r.add_child(comment="just a testing comment in site2")

    def test_for_app_models(self):
        # there is no comment posted yet to article_1 nor article_2
        count = TreeComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 0)
        self.post_comment_1()
        count = TreeComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 1)
        self.post_comment_2()
        count = TreeComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 2)
        self.post_comment_3()
        count = TreeComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 3)
        self.post_comment_4()
        count = TreeComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 4)

    def test_multi_site_for_app_models(self):
        self.post_comment_1()  # To site1.
        self.post_comment_4()  # To site2.
        count_site1 = TreeComment.objects.for_app_models("tests.article",
                                                         site=self.site1).count()
        self.assertEqual(count_site1, 1)
        count_site2 = TreeComment.objects.for_app_models("tests.article",
                                                         site=self.site2).count()
        self.assertEqual(count_site2, 1)


# In order to test methods 'save' and '_calculate_thread_data', simulate the
# following threads, in order of arrival:
#
# testcase cmt.id   parent level-0  level-1  level-2
#  step1     1        -      c1                        <-                 cmt1
#  step1     2        -      c2                        <-                 cmt2
#  step2     3        1      --       c3               <-         cmt1 to cmt1
#  step2     4        1      --       c4               <-         cmt2 to cmt1
#  step3     5        2      --       c5               <-         cmt1 to cmt2
#  step4     6        5      --       --        c6     <- cmt1 to cmt1 to cmt2
#  step4     7        4      --       --        c7     <- cmt1 to cmt2 to cmt1
#  step5     8        3      --       --        c8     <- cmt1 to cmt1 to cmt1
#  step5     9        -      c9                        <-                 cmt9


def thread_test_step_1(article):
    """
    Add Two Comments for the article

    root -
      comment 1
      comment 2
    """
    site = Site.objects.get(pk=1)

    #TreeComment.objects.create()
    # post Comment 1 with parent_id 0
    root = TreeComment.objects.get_or_create_root(article, site=site)

    root.add_child(comment="comment 1 to article",
                   submit_date=datetime.now())
    root.refresh_from_db()

    # post Comment 2 with parent_id 0
    root.add_child(comment="comment 2 to article",
                   submit_date=datetime.now())
    root.refresh_from_db()


def thread_test_step_2(article):
    """
    Add 2 replies to the first comment

    root -
      comment 1
        reply 1
        reply 2
      comment 2

    """
    site = Site.objects.get(pk=1)

    root = TreeComment.objects.get_or_create_root(article, site=site)

    children = root.get_children()
    c1 = children[0]

    # post Comment 3 to parent_id 1
    c1.add_child(comment="comment 1 to comment 1")

    # post Comment 4 to parent_id 1
    c1.add_child(comment="comment 2 to comment 1")


def thread_test_step_3(article):
    """

    root -
      comment 1
        reply 1
        reply 2
      comment 2
        reply 1

    """
    site = Site.objects.get(pk=1)

    root = TreeComment.objects.get_or_create_root(article, site=site)

    c2 = root.get_children()[1]

    c2.add_child(comment="comment 1 to comment 2")


def thread_test_step_4(article):
    """

    root -
      comment 1
        reply 1
        reply 2
            reply 2.1
      comment 2
        reply 1
            reply 1.1

    """
    site = Site.objects.get(pk=1)

    root = TreeComment.objects.get_or_create_root(article, site=site)

    c1, c2 = root.get_children()

    c1_children = c1.get_children()
    c2_children = c2.get_children()

    c2_children[0].add_child(comment="cmt 1 to cmt 1 to cmt 2")

    c1_children[1].add_child(comment="cmt 1 to cmt 2 to cmt 1")


def thread_test_step_5(article):
    """

    root -
      comment 1
        reply 1
        reply 2
            reply 2.1
                reply 2.1.1
      comment 2
        reply 1
            reply 1.1
                reply 1.1.1

    """
    site = Site.objects.get(pk=1)

    root = TreeComment.objects.get_or_create_root(article, site=site)

    c1, c2 = root.get_children()

    c1_children = c1.get_children()
    c2_children = c2.get_children()

    c1_grandchildren_list = [c.get_children() for c in c1_children]
    c2_grandchildren_list = [c.get_children() for c in c2_children]

    for i, gc in enumerate(c1_grandchildren_list):
        for g in gc:
            g.add_child(comment=f'cmt1 to cmt{i} to cmt1 to cmt1')

    for i, gc in enumerate(c2_grandchildren_list):
        for g in gc:
            g.add_child(comment=f'cmt1 to cmt{i} to cmt1 to cmt2')


class BaseThreadStep1TestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        thread_test_step_1(self.article_1)

    def test_threaded_comments_step_1_level_0(self):
        # comment 1
        root = TreeComment.objects.get_root(self.article_1)
        self.assertEqual(root.get_descendant_count(), 2, "Expected to have 2 comments/replies")
        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments, no replies")


class ThreadStep2TestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)

    def test_threaded_comments_step_2_level_0(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        c1, c2 = root.get_children()

        # comment 1
        self.assertEqual(c1.get_children_count(), 2, "Expected comment 1 to have 2 replies")
        # comment 2
        self.assertEqual(c2.get_children_count(), 0, "Expected comment 2 to have no replies")


class ThreadStep3TestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)

    def test_threaded_comments_step_3_level_0(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 5, "Expected to have 5 total comments and replices")


    def test_threaded_comments_step_3_level_1(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 5, "Expected to have 5 total comments and replices")

        c1, c2 = root.get_children()

        # comment 1
        self.assertEqual(c1.get_children_count(), 2, "Expected comment 1 to have 2 replies")
        # comment 2
        self.assertEqual(c2.get_children_count(), 1, "Expected comment 2 to have 1 replie")


class ThreadStep4TestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)
        thread_test_step_4(self.article_1)

    def test_threaded_comments_step_4_level_0(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 7, "Expected to have 7 total comments and replies")

        c1, c2 = root.get_children()

    def test_threaded_comments_step_4_level_1(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 7, "Expected to have 5 total comments and replies")

        c1, c2 = root.get_children()
        self.assertEqual(c1.get_children_count(), 2, "Expected 2 replies to comment 1")
        self.assertEqual(c2.get_children_count(), 1, "Expected 1 reply to comment 2")

        c1_children = c1.get_children()
        c2_children = c2.get_children()
        self.assertEqual(c1_children[1].get_children_count(), 1, "Expected 1 replies to comment 1, reply 2")
        self.assertEqual(c2_children[0].get_children_count(), 1, "Expected 1 reply to comment 2, reply 1")

    def test_threaded_comments_step_4_level_2(self):
        root = TreeComment.objects.get_root(self.article_1)


class ThreadStep5TestCase(ArticleBaseTestCase):
    def setUp(self):
        super().setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)
        thread_test_step_4(self.article_1)
        thread_test_step_5(self.article_1)

    def test_threaded_comments_step_5_level_0(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 9, "Expected to have 5 total comments and replices")

        c1, c2 = root.get_children()

    def test_threaded_comments_step_5_level_1(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 9, "Expected to have 5 total comments and replices")

        c1, c2 = root.get_children()

    def test_threaded_comments_step_5_level_2(self):
        root = TreeComment.objects.get_root(self.article_1)

        self.assertEqual(root.get_children_count(), 2, "Expected to have 2 comments")
        self.assertEqual(root.get_descendant_count(), 9, "Expected to have 5 total comments and replices")

        c1, c2 = root.get_children()

        self.assertEqual(c1.get_children()[1].get_children()[0].get_children_count(),
                         1, "Expected 1 grandchild")
        self.assertEqual(c2.get_children()[0].get_children()[0].get_children_count(),
                         1, "Expected 1 grandchild")

    def test_exceed_max_thread_level_raises_exception(self):
        root = TreeComment.objects.get_or_create_root(self.article_1)
        with self.assertRaises(MaxThreadLevelExceededException):
            comments = root.get_descendants()
            max_depth = [c for c in comments if c.depth > settings.COMMENTS_TREE_MAX_THREAD_LEVEL]
            max_depth[0].add_child(comment="Should cause an exception here")


def add_comment_to_diary_entry(diary):
    root = TreeComment.objects.get_or_create_root(diary)
    root.add_child(comment='This is a diary comment')


class DiaryBaseTestCase(DjangoTestCase):
    def setUp(self):
        self.day_in_diary = Diary.objects.create(body="About Today...")
        self.root = TreeComment.objects.get_or_create_root(self.day_in_diary)

    def test_max_thread_level_by_app_model(self):
        self.root = TreeComment.objects.get_or_create_root(self.day_in_diary)
        c = self.root
        for x in range(1, settings.COMMENTS_TREE_MAX_THREAD_LEVEL+1):
            c = c.add_child(comment=f"Comment level {x}")
        with self.assertRaises(MaxThreadLevelExceededException):
            c.add_child(comment="Comment should cause an exception now")



from typing import Optional, List

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Q
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core import signing
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from django_comments_tree.signals import comment_was_flagged

from django_comments_tree.conf import settings
from treebeard.mp_tree import MP_Node, MP_NodeManager

from .abstract import CommentAbstractModel

LIKEDIT_FLAG = "I liked it"
DISLIKEDIT_FLAG = "I disliked it"


def max_thread_level_for_content_type(content_type):
    app_model = "%s.%s" % (content_type.app_label, content_type.model)
    if app_model in settings.COMMENTS_TREE_MAX_THREAD_LEVEL_BY_APP_MODEL:
        return settings.COMMENTS_TREE_MAX_THREAD_LEVEL_BY_APP_MODEL[app_model]
    else:
        return settings.COMMENTS_TREE_MAX_THREAD_LEVEL


class MaxThreadLevelExceededException(Exception):
    def __init__(self, comment):
        self.comment = comment
        # self.max_by_app = max_thread_level_for_content_type(content_type)

    def __str__(self):
        return "Max thread level reached for comment %d" % self.comment.id


class CommentManager(MP_NodeManager):

    def get_root(self, obj):
        """ Return the root for the given object """
        try:
            ct = ContentType.objects.get_for_model(obj)
            assoc = CommentAssociation.objects.get(content_type=ct, object_id=obj.id)
            return assoc.root
        except ObjectDoesNotExist:
            return None

    def get_or_create_root(self, obj, site=None):
        ct = ContentType.objects.get_for_model(obj)

        if site is None:
            site = Site.objects.get(pk=1)

        try:
            assoc = CommentAssociation.objects.get(content_type=ct,
                                                   object_id=obj.id,
                                                   site=site)
        except ObjectDoesNotExist:
            root = TreeComment.add_root()
            assoc = CommentAssociation.objects.create(content_type=ct,
                                                      object_id=obj.id,
                                                      content_object=obj,
                                                      site=site,
                                                      root=root)
        if assoc:
            return assoc.root
        return None

    def get_or_create_root_type_and_id(self, ct, object_id):
        return None

    def create_for_object(self, obj, comment=''):
        root = self.get_or_create_root(obj)
        return root.add_child(comment=comment)

    def in_moderation(self):
        """
        QuerySet for all comments currently in the moderation queue.
        """
        return self.get_queryset().filter(is_public=False, is_removed=False)

    def for_model(self, model):
        """
        QuerySet for all comments for a particular model (either an instance or
        a class).

        Updated: this can't return a queryset, since MP_Node are queriable that way

        """
        ct = ContentType.objects.get_for_model(model)

        qs = self.get_queryset().filter(content_type=ct)
        if isinstance(model, models.Model):
            qs = qs.filter(object_id=model._get_pk_val())
        return qs

    def for_app_models(self, *args, **kwargs) -> Optional[models.QuerySet]:
        """Return TreeComments for pairs "app.model" given in args"""
        content_types = []
        for app_model in args:
            app, model = app_model.split(".")
            content_types.append(ContentType.objects.get(app_label=app,
                                                         model=model))
        return self.for_content_types(content_types, **kwargs)

    def for_content_types(self,
                          content_types: List[str],
                          site: int = None) -> Optional[models.QuerySet]:
        """
        Return all descendants of the content type.
        :param content_types:
        :param site:
        :return:
        """
        filter_fields = {'content_type__in': content_types}
        if site is not None:
            filter_fields['site'] = site
        associations = CommentAssociation.objects.filter(**filter_fields)
        parent_paths = []

        for assoc in associations:
            parent_paths.append(assoc.root.path)

        Qlist = [Q(path__startswith=path) for path in parent_paths]
        myQ = Qlist.pop()
        for Qitem in Qlist:
            myQ |= Qitem

        qs = TreeComment.objects.filter(depth__gte=2).filter(myQ)
        return qs

    def count_for_content_types(self, content_types: List[str], site: int = None) -> int:
        count = 0
        filter_fields = {'content_type__in': content_types}
        if site is not None:
            filter_fields['site'] = site
        associations = CommentAssociation.objects.filter(**filter_fields)
        for assoc in associations:
            count += assoc.root.get_descendants().count()
        return count

    def get_queryset(self):
        qs = super().get_queryset()
        return qs


class CommentAssociation(models.Model):
    """
    Associate a tree node with a particular model by GenericForeignKey

    ToDo: Review the proper way to use GFK's. Do I need all of the other parts?
    """

    # Root of comments for the associated model
    root = models.ForeignKey('TreeComment', on_delete=models.CASCADE, null=True)

    # Content-object field
    content_type = models.ForeignKey(ContentType,
                                     verbose_name=_('content type'),
                                     related_name="content_type_set_for_%(class)s",
                                     on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey(ct_field="content_type", fk_field="object_id")

    # Retained a legacy. Remove once I determine it is not needed
    # object_pk = models.TextField(_('object ID'))

    # Metadata about the comment
    # ToDo: Why do I need this?
    site = models.ForeignKey(Site, on_delete=models.CASCADE)

    @property
    def object_pk(self):
        return str(self.object_id)


class TreeComment(MP_Node, CommentAbstractModel):
    node_order_by = ['submit_date']

    def __init__(self, *args, **kwargs):
        self._association = None
        super().__init__(*args, **kwargs)

    followup = models.BooleanField(blank=True, default=False,
                                   help_text=_("Notify follow-up comments"))
    objects = CommentManager()

    def add_child(self, *args, comment=None, **kwargs):
        """ Check for maximum depth before adding child """
        depth = self.depth
        if depth > settings.COMMENTS_TREE_MAX_THREAD_LEVEL:
            raise MaxThreadLevelExceededException(comment)

        if 'instance' in kwargs:
            child = super().add_child(*args, **kwargs)
        else:
            child = super().add_child(*args, comment=comment, **kwargs)
        return child

    def save(self, *args, **kwargs):
        is_new = self.pk is None
        super().save(*args, **kwargs)
        if is_new:
            if not self.is_root():
                root = self.get_root()
                try:
                    assoc = CommentAssociation.objects.get(root=root)
                    max_level = max_thread_level_for_content_type(assoc.content_type)
                    if max_level and self.get_depth() > max_level:
                        raise MaxThreadLevelExceededException(self)
                except Exception:
                    pass
            kwargs["force_insert"] = False
            super().save(*args, **kwargs)

    def get_reply_url(self):
        return reverse("comments-tree-reply", kwargs={"cid": self.pk})

    def allow_thread(self):
        if self.get_depth() < max_thread_level_for_content_type(self.content_type):
            return True
        else:
            return False

    @property
    def association(self):
        """
        Return the content type for this comment. We have to search from the root for this.
        Cache the value though.
        """
        if self._association is None:
            root = self if self.is_root() else self.get_root()
            self._association = CommentAssociation.objects.get(root=root)

        return self._association

    @property
    def content_type(self):
        """
        Return the content type for this comment. We have to search from the root for this.
        Cache the value though.
        """
        assoc = self.association
        if assoc:
            return assoc.content_type

        return None

    @property
    def object_id(self):
        """ Accessor added for compatibility with django_contrib.comments """
        assoc = self.association
        if assoc:
            return assoc.object_id

        return None

    @property
    def site(self):
        """ Accessor added for compatibility with django_contrib.comments """
        assoc = self.association
        if assoc:
            return assoc.site

        return None

    @classmethod
    def tree_from_comment(cls, root,
                          filter_public=True,):
        """
        Return a recursive structure with comments and their children,
        starting at the given root.
        """
        retval = []
        children = root.get_children().order_by('submit_date')
        if filter_public:
            children = children.filter(is_public=True)
        for child in children:
            data = {
                "comment": child,
                "children": cls.tree_from_comment(
                    child,
                    filter_public=filter_public
                )
            }
            retval.append(data)
        return retval

    @classmethod
    def tree_for_associated_object(cls, obj,
                                   with_flagging=False,
                                   with_feedback=False,
                                   user=None):

        root = TreeComment.objects.get_or_create_root(obj)
        data = cls.tree_from_comment(root)
        return data

    def users_flagging(self, flag):
        return [obj.user for obj in self.flags.filter(flag=flag)]


@receiver(comment_was_flagged)
def unpublish_nested_comments_on_removal_flag(sender, comment, flag, **kwargs):
    if flag.flag == TreeCommentFlag.MODERATOR_DELETION:
        TreeComment.objects.filter(~(Q(pk=comment.id)), parent_id=comment.id) \
            .update(is_public=False)


class DummyDefaultManager:
    """
    Dummy Manager to mock django's CommentForm.check_for_duplicate method.
    """

    def __getattr__(self, name):
        return lambda *args, **kwargs: []

    def using(self, *args, **kwargs):
        return self


class TmpTreeComment(dict):
    """
    Temporary TreeComment to be pickled, zipped and appended to a URL.
    """
    _default_manager = DummyDefaultManager()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            if self.get('tree_comment'):
                try:
                    return getattr(self.get('tree_comment'), key)
                except KeyError:
                    pass
                except Exception:
                    pass
            return None

    def __setattr__(self, key, value):
        self[key] = value

    def save(self, *args, **kwargs):
        pass

    def _get_pk_val(self):
        if self.tree_comment:
            return self.tree_comment._get_pk_val()
        else:
            content_type = "%s.%s" % self.content_type.natural_key()
            return signing.dumps("%s:%s" % (content_type, self.object_id))

    def __setstate__(self, state):
        ct_key = state.pop('content_type_key')
        ctype = ContentType.objects.get_by_natural_key(*ct_key)
        self.update(
            state,
            content_type=ctype,
            content_object=ctype.get_object_for_this_type(
                pk=state['object_id']
            )
        )

    def __reduce__(self):
        state = {k: v for k, v in self.items() if k != 'content_object'}
        ct = state.pop('content_type')
        state['content_type_key'] = ct.natural_key()
        return (TmpTreeComment, (), state,)


# ----------------------------------------------------------------------
class BlackListedDomain(models.Model):
    """
    A blacklisted domain from which comments should be discarded.
    Automatically populated with a small amount of spamming domains,
    gathered from http://www.joewein.net/spam/blacklist.htm

    You can download for free a recent version of the list, and subscribe
    to get notified on changes. Changes can be fetched with rsync for a
    small fee (check their conditions, or use any other Spam filter).
    """
    domain = models.CharField(max_length=200, db_index=True)

    def __str__(self):
        return self.domain

    class Meta:
        ordering = ('domain',)


class TreeCommentFlag(models.Model):
    """
    Records a flag on a comment. This is intentionally flexible; right now, a
    flag could be:

        * A "removal suggestion" -- where a user suggests a comment for (potential) removal.

        * A "moderator deletion" -- used when a moderator deletes a comment.

    You can (ab)use this model to add other flags, if needed. However, by
    design users are only allowed to flag a comment with a given flag once;
    if you want rating look elsewhere.
    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_('user'), related_name="treecomment_flags",
        on_delete=models.CASCADE,
    )
    comment = models.ForeignKey(
        # Translators: 'comment' is a noun here.
        TreeComment, verbose_name=_('comment'), related_name="flags", on_delete=models.CASCADE,
    )
    # Translators: 'flag' is a noun here.
    flag = models.CharField(_('flag'), max_length=30, db_index=True)
    flag_date = models.DateTimeField(_('date'), default=None)

    # Constants for flag types
    SUGGEST_REMOVAL = "removal suggestion"
    MODERATOR_DELETION = "moderator deletion"
    MODERATOR_APPROVAL = "moderator approval"

    class Meta:
        unique_together = [('user', 'comment', 'flag')]
        verbose_name = _('comment flag')
        verbose_name_plural = _('comment flags')

    def __str__(self):
        return "%s flag of comment ID %s by %s" % (
            self.flag, self.comment_id, self.user.get_username()
        )

    def save(self, *args, **kwargs):
        if self.flag_date is None:
            self.flag_date = timezone.now()
        super().save(*args, **kwargs)
